from __future__ import annotations

from collections.abc import Mapping

import jax.numpy as jnp
import jax.random as jr

from .annotations import JaxIntegralArray, KeyArray


class RngStream:
    def __init__(self, key: KeyArray, count: JaxIntegralArray | None = None) -> None:
        super().__init__()
        if count is None:
            count = jnp.zeros((), dtype=jnp.uint32)
        else:
            assert count.ndim == 0
        self._key = key
        self.count = count

    def key(self) -> KeyArray:
        key = jr.fold_in(self._key, self.count)
        self.count += 1
        return key


def create_streams(keys: Mapping[str, KeyArray]) -> Mapping[str, RngStream]:
    return {name: RngStream(key) for name, key in keys.items()}


def sample_streams(rngs: Mapping[str, RngStream]) -> Mapping[str, KeyArray]:
    return {name: stream.key() for name, stream in rngs.items()}


def fork_streams(rngs: Mapping[str, RngStream], samples: int) -> Mapping[str, KeyArray]:
    return {name: jr.split(stream.key(), samples) for name, stream in rngs.items()}



from __future__ import annotations

import math
import operator
from typing import Any

import jax
import jax.numpy as jnp

from .annotations import JaxArray, JaxBooleanArray


def dynamic_tree_all(tree: Any) -> JaxBooleanArray:
    """Like `jax.tree.all`, but can be used in dynamic code like jitted functions and loops."""
    return jax.tree.reduce(jnp.logical_and, tree, jnp.asarray(True))


def tree_sum(x: Any) -> JaxArray:
    if not jax.tree.leaves(x):
        return jnp.zeros(())
    retval = jax.tree.reduce(operator.add, jax.tree.map(jnp.sum, x))
    assert isinstance(retval, JaxArray)
    assert retval.ndim == 0
    return retval


def element_count(x: Any) -> int:
    def array_element_count(x: Any) -> int:
        if not isinstance(x, JaxArray):
            raise TypeError
        return math.prod(x.shape)
    retval = jax.tree.reduce(operator.add, jax.tree.map(array_element_count, x), 0)
    assert isinstance(retval, int)
    return retval



from ._src.dataclasses.dataclass import DataclassInstance, TDataclassInstance, dataclass
from ._src.dataclasses.helpers import as_shallow_dict, field

__all__ = ['DataclassInstance', 'TDataclassInstance', 'as_shallow_dict', 'dataclass', 'field']

try:
    import flax
except ImportError:
    pass
else:
    flax_version = tuple(int(x) for x in flax.__version__.split('.'))
    if flax_version > (0, 10):
        from ._src.dataclasses.flax import DataClassModule, module_field
        __all__ += ['DataClassModule', 'module_field']



from __future__ import annotations

import dataclasses
from collections.abc import Callable, Mapping
from dataclasses import MISSING, fields
from typing import Any, TypeVar, overload

T = TypeVar('T')


# NOTE: Actual return type is 'Field[T]', but we want to help type checkers
# to understand the magic that happens at runtime.
@overload  # `default` and `default_factory` are optional and mutually exclusive.
def field(*, static: bool = False, default: T, init: bool = ...,
          repr: bool = ...,
          hash: bool | None = ...,
          compare: bool = ...,
          metadata: Mapping[str, Any] | None = ..., kw_only: bool = ...) -> T:
    ...


@overload
def field(*, static: bool = False, default_factory: Callable[[], T], init: bool = ...,
          repr: bool = ...,
          hash: bool | None = ...,
          compare: bool = ...,
          metadata: Mapping[str, Any] | None = ..., kw_only: bool = ...) -> T:
    ...


@overload
def field(*, static: bool = False, init: bool = ...,
          repr: bool = ...,
          hash: bool | None = ...,
          compare: bool = ...,
          metadata: Mapping[str, Any] | None = ..., kw_only: bool = ...) -> Any:
    ...


def field(*, static: bool = False, default: Any = MISSING,
          default_factory: Any = MISSING, init: bool = True,
          repr: bool = True,  # noqa: A002
          hash: bool | None = None,  # noqa: A002
          compare: bool = True,
          metadata: Mapping[str, Any] | None = None,
          kw_only: Any = MISSING) -> Any:
    """A field creator with a static indicator.

    The static flag indicates whether a field is a pytree or static.  Pytree fields are
    differentiated and traced.  Static fields are hashed and compared.
    """
    metadata_dict: dict[str, Any] = {} if metadata is None else dict(metadata)
    metadata_dict['static'] = static
    if default is MISSING:
        return dataclasses.field(default_factory=default_factory,
                                 init=init,
                                 repr=repr,
                                 hash=hash,
                                 compare=compare,
                                 metadata=metadata_dict,
                                 kw_only=kw_only)
    return dataclasses.field(default=default,
                             init=init,
                             repr=repr,
                             hash=hash,
                             compare=compare,
                             metadata=metadata_dict,
                             kw_only=kw_only)


def as_shallow_dict(dcls: Any) -> dict[str, Any]:
    return {field.name: getattr(dcls, field.name) for field in fields(dcls)}



from __future__ import annotations

import operator
from collections.abc import Callable
from functools import partial
from typing import Any, TypeVar, cast

import jax.numpy as jnp
from jax import tree, vjp
from jax.custom_derivatives import zero_from_primal as jax_zero_from_primal

from .annotations import JaxRealArray, RealNumeric
from .display.print_generic import print_generic
from .shims import custom_jvp, custom_vjp

X = TypeVar('X')
XT = TypeVar('XT', bound=tuple[Any, ...])
Y = TypeVar('Y')


def zero_from_primal(x: X, /, *, symbolic_zeros: bool = False) -> X:
    return jax_zero_from_primal(x, symbolic_zeros=symbolic_zeros)


# scale_cotangent ----------------------------------------------------------------------------------
def scale_cotangent(x: X,
                    scalar_scale: RealNumeric | None = None,
                    tree_scale: X | None = None,
                    ) -> X:
    """Scale x's cotangent.

    Args:
        x: The principal value.
        scalar_scale: A scalar that scales x's cotangent.
        tree_scale: A tree matching x's shape that scales x's cotangent.
    """
    return x


def _scale_cotangent_jvp(scalar_scale: RealNumeric | None,
                         tree_scale: X | None,
                         primals: tuple[X],
                         tangents: tuple[X]
                         ) -> tuple[X, X]:
    x, = primals
    x_dot, = tangents
    if scalar_scale is not None:
        x_dot = tree.map(lambda x_dot_i: x_dot_i * scalar_scale, x_dot)
    if tree_scale is not None:
        x_dot = tree.map(operator.mul, x_dot, tree_scale)
    return x, x_dot


scale_cotangent = custom_jvp(scale_cotangent, nondiff_argnums=(1, 2))
scale_cotangent.defjvp(_scale_cotangent_jvp)


# reverse_scale_cotangent --------------------------------------------------------------------------
@custom_vjp
def reverse_scale_cotangent(x: X) -> tuple[X, JaxRealArray]:
    """Output x and a dummy zero; scale x's cotangent by the zero's cotangent."""
    return x, jnp.zeros(())


def _reverse_scale_cotangent_fwd(x: X) -> tuple[tuple[X, JaxRealArray], None]:
    return (x, jnp.zeros(())), None


def _reverse_scale_cotangent_bwd(residuals: None, xy_bar: tuple[X, JaxRealArray]) -> tuple[X]:
    del residuals
    x_bar, y_bar = xy_bar
    return (tree.map(lambda x_bar_i: x_bar_i * y_bar, x_bar),)


reverse_scale_cotangent.defvjp(_reverse_scale_cotangent_fwd, _reverse_scale_cotangent_bwd)


# replace_cotangent --------------------------------------------------------------------------------
@custom_vjp
def replace_cotangent(x: X, new_cotangent: X) -> X:
    """Set x's cotangent to be new_cotangent's primal value."""
    assert tree.structure(x) == tree.structure(new_cotangent)
    return x


def _replace_cotangent_fwd(x: X, new_cotangent: X) -> tuple[X, X]:
    assert tree.structure(x) == tree.structure(new_cotangent)
    return x, new_cotangent


def _replace_cotangent_bwd(residuals: X, x_bar: X) -> tuple[X, X]:
    assert tree.structure(residuals) == tree.structure(x_bar)
    return residuals, x_bar


replace_cotangent.defvjp(_replace_cotangent_fwd, _replace_cotangent_bwd)


# copy_cotangent -----------------------------------------------------------------------------------
@custom_vjp
def copy_cotangent(x: X, y: X) -> X:
    """Output x, and copy its cotangent to y."""
    assert tree.structure(x) == tree.structure(y)
    return x


def _copy_cotangent_fwd(x: X, y: X) -> tuple[X, None]:
    assert tree.structure(x) == tree.structure(y)
    return x, None


def _copy_cotangent_bwd(residuals: None, x_bar: X) -> tuple[X, X]:
    del residuals
    return x_bar, x_bar


copy_cotangent.defvjp(_copy_cotangent_fwd, _copy_cotangent_bwd)


# print_cotangent ----------------------------------------------------------------------------------
@partial(custom_vjp, static_argnums=(1,))
def print_cotangent(u: X, name: str | None = None) -> X:
    """Print the cotangent of u."""
    return u


def _print_cotangent_fwd(u: X, name: str | None) -> tuple[X, None]:
    return u, None


def _print_cotangent_bwd(name: str | None, residuals: None, x_bar: X) -> tuple[X]:
    del residuals
    if name is None:
        print_generic(x_bar)
    else:
        print_generic(**{name: x_bar})  # type: ignore[arg-type] # pyright: ignore
    return (x_bar,)


# https://github.com/python/mypy/issues/14802
print_cotangent.defvjp(_print_cotangent_fwd,  # type: ignore[arg-type]
                       _print_cotangent_bwd)


# cotangent_combinator -----------------------------------------------------------------------------
def cotangent_combinator(f: Callable[..., tuple[XT, Y]],
                         args_tuples: tuple[tuple[Any, ...], ...],
                         aux_cotangent_scales: tuple[RealNumeric, ...] | None) -> tuple[XT, Y]:
    """Run a function once, but send differerent cotangents back to each input.

    Args:
        f: A function that accepts positional arguments and returns xs, y where xs is a tuple of
            length n.
        args_tuples: n copies of the same tuple of positional arguments accepted by f.
        aux_cotangent_scales: If provided, scale each of the cotangents.
    Returns: The pair (xs, y).

    The purpose of the cotangent combinator is to take cotangents of each of the elements of x, and
    send them back through to each of the corresponding argument tuples.
    """
    return f(*args_tuples[0])


cotangent_combinator = custom_vjp(cotangent_combinator, static_argnums=(0, 2))


def _cotangent_combinator_fwd(f: Callable[..., tuple[XT, Y]],
                              args_tuples: tuple[tuple[Any, ...], ...],
                              aux_cotangent_scales: tuple[RealNumeric, ...] | None
                              ) -> tuple[tuple[XT, Y],
                                         Callable[[tuple[XT, Y]], tuple[Any, ...]]]:
    return vjp(f, *args_tuples[0])


def _cotangent_combinator_bwd(f: Callable[..., tuple[XT, Y]],
                              aux_cotangent_scales: tuple[RealNumeric, ...] | None,
                              f_vjp: Callable[[tuple[XT, Y]], tuple[Any, ...]],
                              xy_bar: tuple[XT, Y]
                              ) -> tuple[Any, ...]:
    xs_bar, y_bar = xy_bar
    if aux_cotangent_scales is None:
        aux_cotangent_scales = tuple(1.0 for _ in xs_bar)
    xs_zero = tuple(tree.map(jnp.zeros_like, x_bar)
                    for x_bar in xs_bar)
    all_args_bar = []
    for i, (x_bar, aux_cotangent_scale) in enumerate(zip(xs_bar, aux_cotangent_scales,
                                                         strict=True)):
        scaled_y_bar = tree.map(lambda y_bar_i, scale=aux_cotangent_scale: y_bar_i * scale,
                                y_bar)
        this_xs_bar = cast('XT', (*xs_zero[:i], x_bar, *xs_zero[i + 1:]))
        this_result_bar = (this_xs_bar, scaled_y_bar)
        this_args_bar = f_vjp(this_result_bar)
        all_args_bar.append(this_args_bar)
    return (tuple(all_args_bar),)


cotangent_combinator.defvjp(_cotangent_combinator_fwd, _cotangent_combinator_bwd)



from __future__ import annotations

from functools import partial

import jax.numpy as jnp
import jax.random as jr
from jax import jit
from jax.lax import scan
from jax.ops import segment_sum

from .annotations import JaxArray, KeyArray
from .dataclasses.dataclass import dataclass

__all__ = ['cluster_with_k_means']


_k_means_dtype = jnp.int32


@partial(jit, static_argnames=('n_clusters',))
def _kmeans_plus_plus_init(key: KeyArray, points: JaxArray, n_clusters: int) -> JaxArray:
    # Initialize the centroid array.
    n_samples, n_features = points.shape
    centroids = jnp.empty((n_clusters, n_features))
    keys = jr.split(key, n_clusters)

    # Choose the first centroid randomly.
    first_idx = jr.randint(keys[0], shape=(), minval=0, maxval=n_samples)
    centroids = centroids.at[0].set(points[first_idx])

    for i, key_i in enumerate(keys[1:]):
        # Compute all distances between all centroids and all points.
        # Has shape (n_samples, n_clusters, n_features).
        distances = jnp.square(points[:, jnp.newaxis, :] - centroids[jnp.newaxis, :, :])
        distances = jnp.sum(distances, axis=2)  # Has shape (n_samples, n_clusters).

        # Compute distance to closest centroid from all points.
        distances = jnp.min(distances, axis=1)  # Has shape (n_samples,).

        # Sample a new centroid with probability proportional to the distances.
        p = distances / jnp.sum(distances)
        next_idx = jr.choice(key_i, a=n_samples, p=p)
        centroids = centroids.at[i].set(points[next_idx])
    return centroids


@dataclass
class _KMeansCarry:
    centroids: JaxArray
    cluster_assignments: JaxArray


def _k_means_single_iteration(points: JaxArray,
                              n_clusters: int,
                              carry: _KMeansCarry,
                              _: None
                              ) -> tuple[_KMeansCarry, None]:
    # Has shape (n_samples, n_clusters, n_features).
    distances = jnp.square(points[:, jnp.newaxis] - carry.centroids[jnp.newaxis])
    distances = jnp.sum(distances, axis=2)  # Has shape (n_samples, n_clusters).
    # Has shape (n_samples,)
    best_centroids = jnp.argmin(distances, axis=1).astype(_k_means_dtype)

    # total = jnp.zeros(n_clusters).at[best_centroids].add(points)
    # count = jnp.zeros(n_clusters).at[best_centroids].add(1.0)
    total = segment_sum(points, best_centroids, n_clusters)  # (n_clusters, n_features)
    count = segment_sum(1.0, best_centroids, n_clusters)  # (n_clusters,)
    centroids = total / count[slice(None), jnp.newaxis]
    return _KMeansCarry(centroids, best_centroids), None


@partial(jit, static_argnames=('iterations',))
def _k_means_many_iterations(points: JaxArray,
                             initial_means: JaxArray,
                             iterations: int
                             ) -> tuple[JaxArray, JaxArray]:
    n_samples = points.shape[0]
    n_clusters = initial_means.shape[0]
    cluster_assignments = jnp.zeros(n_samples, dtype=_k_means_dtype)
    carry = _KMeansCarry(initial_means, cluster_assignments)
    k_means_update = partial(_k_means_single_iteration, points, n_clusters)
    carry, _ = scan(k_means_update, carry, None, iterations)
    return carry.centroids, carry.cluster_assignments


def cluster_with_k_means(key: KeyArray,
                         points: JaxArray,
                         n_clusters: int,
                         iterations: int
                         ) -> JaxArray:
    """Cluster with k-means.

    Args:
        key: A random key.
        points: An array with shape (n_samples, n_features).
        n_clusters: The number of clusters to produce.
        iterations: The number of iterations to do.
    """
    initial_means = _kmeans_plus_plus_init(key, points, n_clusters)
    _, cluster_assignments = _k_means_many_iterations(points, initial_means, iterations)
    return cluster_assignments



from __future__ import annotations

from collections.abc import Callable
from dataclasses import KW_ONLY, asdict
from typing import Any, Generic

import jax.numpy as jnp
from optax import (GradientTransformationExtraArgs, adabelief, adadelta, adafactor, adagrad, adam,
                   adamax, adamaxw, adamw, fromage, lamb, lars, lbfgs, lion, noisy_sgd, novograd,
                   optimistic_gradient_descent, polyak_sgd, radam, rmsprop,
                   scale_by_zoom_linesearch, sgd, sm3, yogi)
from optax.contrib import dpsgd
from typing_extensions import override

from tjax.dataclasses import dataclass, field

from ..annotations import IntegralNumeric, JaxArray, RealNumeric
from .transform import GenericGradientState, GradientTransformation, Weights
from .transforms import Schedule

# Types --------------------------------------------------------------------------------------------
ScalarOrSchedule = float | JaxArray | Schedule


# Transforms from optax._src.alias.py --------------------------------------------------------------
@dataclass
class AdaBelief(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-8
    eps_root: RealNumeric = 1e-16

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(adabelief(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *adabelief(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class AdaDelta(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule | None = None
    rho: RealNumeric = 0.9
    eps: RealNumeric = 1e-6
    weight_decay: RealNumeric = 0.0
    weight_decay_mask: bool | Weights | Callable[[Weights], Any] | None = field(default=None,
                                                                                static=True)

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(adadelta(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *adadelta(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class AdaFactor(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule | None = None
    min_dim_size_to_factor: IntegralNumeric = 128
    decay_rate: RealNumeric = 0.8
    decay_offset: IntegralNumeric = 0
    multiply_by_parameter_scale: RealNumeric = True
    clipping_threshold: float | None = 1.0
    momentum: float | None = None
    dtype_momentum: Any = field(default=jnp.float32, static=True)
    weight_decay_rate: float | None = None
    eps: RealNumeric = 1e-30
    factored: bool = field(default=True, static=True)
    weight_decay_mask: bool | Weights | Callable[[Weights], Any] | None = field(default=None,
                                                                                static=True)

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(adafactor(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *adafactor(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class AdaGrad(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    initial_accumulator_value: RealNumeric = 0.1
    eps: RealNumeric = 1e-7

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(adagrad(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *adagrad(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class Adam(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-8
    eps_root: RealNumeric = 0.0
    mu_dtype: Any | None = field(default=None, static=True)
    _: KW_ONLY
    nesterov: bool = field(default=False, static=True)

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(adam(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *adam(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class AdamW(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: RealNumeric
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-8
    eps_root: RealNumeric = 0.0
    mu_dtype: Any | None = field(default=None, static=True)
    weight_decay: RealNumeric = 1e-4
    mask: Any | Callable[[Any], Any] | None = None
    _: KW_ONLY
    nesterov: bool = field(default=False, static=True)

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(adamw(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *adamw(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class Adamax(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: RealNumeric
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-8

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(adamax(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *adamax(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class AdamaxW(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: RealNumeric
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-8
    weight_decay: RealNumeric = 1e-4
    mask: Any | Callable[[Any], Any] | None = None

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(adamaxw(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *adamaxw(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class Fromage(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: RealNumeric
    min_norm: RealNumeric = 1e-6

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(fromage(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *fromage(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class Lamb(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-6
    eps_root: RealNumeric = 0.0
    weight_decay: RealNumeric = 0.0
    mask: bool | Weights | Callable[[Weights], Any] | None = field(default=None, static=True)

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(lamb(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *lamb(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class LARS(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    weight_decay: RealNumeric = 0.0
    weight_decay_mask: bool | Weights | Callable[[Weights], Any] | None = field(default=True,
                                                                                static=True)
    trust_coefficient: RealNumeric = 0.001
    eps: RealNumeric = 0.0
    trust_ratio_mask: bool | Weights | Callable[[Weights], Any] | None = field(default=True,
                                                                               static=True)
    momentum: RealNumeric = 0.9
    nesterov: bool = field(default=False, static=True)

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(lars(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *lars(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class LBFGS(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule | None = None
    memory_size: int = field(default=10, static=True)
    scale_init_precond: bool = field(default=True, static=True)
    linesearch: GradientTransformationExtraArgs | None = (
            scale_by_zoom_linesearch(max_linesearch_steps=15))

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(lars(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        # TODO: The update needs parameters.
        return GenericGradientState.wrap(  # pyright: ignore
            *lbfgs(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class Lion(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.99
    mu_dtype: Any | None = field(default=None, static=True)
    weight_decay: RealNumeric = 1e-3
    mask: Any | Callable[[Any], Any] | None = None

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(lion(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *lion(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class NoisySGD(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    eta: RealNumeric = 0.01
    gamma: RealNumeric = 0.55
    seed: IntegralNumeric = 0

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(noisy_sgd(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *noisy_sgd(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class Novograd(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.25
    eps: RealNumeric = 1e-6
    eps_root: RealNumeric = 0.0
    weight_decay: RealNumeric = 0.0

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(novograd(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *novograd(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class OptimisticGradientDescent(GradientTransformation[GenericGradientState, Weights],
                                Generic[Weights]):
    learning_rate: ScalarOrSchedule
    alpha: ScalarOrSchedule
    beta: ScalarOrSchedule

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(optimistic_gradient_descent(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *optimistic_gradient_descent(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class PolyakSGD(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    max_learning_rate: RealNumeric = 1.0
    scaling: ScalarOrSchedule = 1.0
    f_min: RealNumeric = 0.0
    eps: RealNumeric = 0.0

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(polyak_sgd(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        # TODO: The update needs parameters.
        return GenericGradientState.wrap(  # pyright: ignore
            *polyak_sgd(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class RAdam(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-8
    eps_root: RealNumeric = 0.0
    threshold: RealNumeric = 5.0
    _: KW_ONLY
    nesterov: bool = field(default=False, static=True)

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(radam(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *radam(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class RMSProp(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    """A flexible RMSProp optimiser.

    RMSProp is an SGD variant with learning rate adaptation. The `learning_rate`
    used for each weight is scaled by a suitable estimate of the magnitude of the
    gradients on previous steps. Several variants of RMSProp can be found
    in the literature. This alias provides an easy to configure RMSProp
    optimiser that can be used to switch between several of these variants.

    References:
        Tieleman and Hinton, 2012: http://www.cs.toronto.edu/~hinton/coursera/lecture6/lec6.pdf
        Graves, 2013: https://arxiv.org/abs/1308.0850
    Args:
        learning_rate: this is a fixed global scaling factor.
        decay: the decay used to track the magnitude of previous gradients.
        eps: a small numerical constant to avoid dividing by zero when rescaling.
        initial_scale: (default `0.`), initialisation of accumulators tracking the
        magnitude of previous updates. PyTorch uses `0`, TF1 uses `1`. When
        reproducing results from a paper, verify the value used by the authors.
        centered: (default `False`), whether the second moment or the variance of
        the past gradients is used to rescale the latest gradients.
        momentum: (default `None`), the `decay` rate used by the momentum term,
        when it is set to `None`, then momentum is not used at all.
        nesterov (default `False`): whether nesterov momentum is used.
    """
    learning_rate: RealNumeric
    decay: RealNumeric = 0.9
    eps: RealNumeric = 1e-8
    initial_scale: RealNumeric = 0.0
    centered: bool = field(default=False, static=True)
    momentum: float | None = None
    nesterov: bool = field(default=False, static=True)
    bias_correction: bool = field(default=False, static=True)

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(rmsprop(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *rmsprop(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class SGD(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    """A canonical Stochastic Gradient Descent optimiser.

    This implements stochastic gradient descent. It also includes support for
    momentum, and nesterov acceleration, as these are standard practice when
    using stochastic gradient descent to train deep neural networks.

    References:
        Sutskever et al, 2013: http://proceedings.mlr.press/v28/sutskever13.pdf

    Args:
        learning_rate: this is a fixed global scaling factor.
        momentum: (default `None`), the `decay` rate used by the momentum term,
        when it is set to `None`, then momentum is not used at all.
        nesterov (default `False`): whether nesterov momentum is used.
        accumulator_dtype: optional `dtype` to be used for the accumulator; if
        `None` then the `dtype` is inferred from `params` and `updates`.

    Returns:
        A `GradientTransformation`.
    """
    learning_rate: RealNumeric
    momentum: float | None = None
    nesterov: bool = field(default=False, static=True)
    accumulator_dtype: Any | None = field(default=None, static=True)

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(sgd(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *sgd(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class SM3(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: RealNumeric
    momentum: RealNumeric = 0.9

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(sm3(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *sm3(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class Yogi(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-3

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(yogi(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *yogi(**asdict(self)).update(gradient, state.data, parameters))


@dataclass
class DPSGD(GradientTransformation[GenericGradientState, Weights], Generic[Weights]):
    learning_rate: ScalarOrSchedule
    l2_norm_clip: RealNumeric
    noise_multiplier: RealNumeric
    seed: IntegralNumeric
    momentum: float | None = None
    nesterov: bool = field(default=False, static=True)

    @override
    def init(self, parameters: Weights
             ) -> GenericGradientState:
        return GenericGradientState(dpsgd(**asdict(self)).init(parameters))

    @override
    def update(self,
               gradient: Weights,
               state: GenericGradientState,
               parameters: Weights | None
               ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore
            *dpsgd(**asdict(self)).update(gradient, state.data, parameters))



from __future__ import annotations

from typing import overload

import jax.numpy as jnp
import jax.random as jr
from array_api_compat import array_namespace

from .annotations import (ComplexArray, IntegralArray, JaxComplexArray, JaxIntegralArray,
                          JaxRealArray, KeyArray, RealArray)


@overload
def leaky_integrate(value: JaxComplexArray,
                    time_step: JaxRealArray,
                    drift: JaxComplexArray | None = None,
                    decay: JaxComplexArray | None = None,
                    *,
                    leaky_average: bool = False
                    ) -> JaxComplexArray:
    ...


@overload
def leaky_integrate(value: ComplexArray,
                    time_step: RealArray,
                    drift: ComplexArray | None = None,
                    decay: ComplexArray | None = None,
                    *,
                    leaky_average: bool = False
                    ) -> ComplexArray:
    ...


def leaky_integrate(value: ComplexArray,
                    time_step: RealArray,
                    drift: ComplexArray | None = None,
                    decay: ComplexArray | None = None,
                    *,
                    leaky_average: bool = False
                    ) -> ComplexArray:
    """Update the value so that it is the leaky integral (or leaky average).

    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        decay: If provided, must have positive real component, and the value decays by exp(-decay)
            every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    if drift is None:
        if decay is None:
            return value
        xp = array_namespace(decay, time_step, value)
        return xp.exp(-decay * time_step) * value

    if decay is None:
        if leaky_average:
            raise ValueError
        xp = array_namespace(drift, time_step, value)
        return xp.asarray(value + drift * time_step)

    xp = array_namespace(drift, decay, time_step, value)
    scaled_integrand = (drift / decay) * -xp.expm1(-decay * time_step)

    if leaky_average:
        scaled_integrand *= decay.real

    return xp.exp(-decay * time_step) * value + scaled_integrand


@overload
def leaky_data_weight(iterations_times_time_step: JaxRealArray | JaxIntegralArray,
                      decay: JaxRealArray
                      ) -> JaxRealArray:
    ...


@overload
def leaky_data_weight(iterations_times_time_step: RealArray | IntegralArray,
                      decay: RealArray
                      ) -> RealArray:
    ...


def leaky_data_weight(iterations_times_time_step: RealArray | IntegralArray,
                      decay: RealArray
                      ) -> RealArray:
    """The amount of data that has been incorporated and has not been decayed.

    This equals leaky_integrate(0.0, iterations_times_time_step, 1.0, decay, leaky_average=True).
    """
    xp = array_namespace(iterations_times_time_step, decay)
    return -xp.expm1(-iterations_times_time_step * decay)


def diffused_leaky_integrate(value: JaxComplexArray,
                             time_step: JaxRealArray,
                             rng: KeyArray,
                             diffusion: JaxRealArray,
                             drift: JaxComplexArray | None = None,
                             decay: JaxComplexArray | None = None,
                             *,
                             leaky_average: bool = False
                             ) -> ComplexArray:
    """Update an Ornstein-Uhlenbeck process.

    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        rng: The key array for the stochastic process.
        diffusion: The diffusion for the stochastic process.
        decay: If provided, the value decays by exp(-decay) every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    variance = (diffusion * time_step
                if decay is None
                else diffusion / decay * -jnp.expm1(-decay * time_step))
    jump = jnp.sqrt(variance) * jr.normal(rng, value.shape)
    return leaky_integrate(value, time_step, drift, decay, leaky_average=leaky_average) + jump



import math
from functools import partial
from typing import Any

import jax.numpy as jnp
import pytest
from jax import vjp
from numpy.testing import assert_allclose

from tjax import Array, JaxRealArray, divide_where, normalize


@pytest.mark.parametrize(("x", "axis", "result"),
                         [((1.0, 2.0, 3.0, 4.0), 0, (0.1, 0.2, 0.3, 0.4)),
                          ((0.0, 0.0), 0, (0.5, 0.5)),
                          ])
def test_l1(x: Any,
            axis: tuple[int, ...] | int | None,
            result: JaxRealArray,
            ) -> None:
    assert_allclose(normalize('l1', jnp.asarray(x), axis=axis), jnp.asarray(result))


@pytest.mark.parametrize(("x", "axis", "result"),
                         [((5.0, 12.0), 0, (5 / 13, 12 / 13)),
                          ((0.0, 0.0), 0, jnp.ones(2) / jnp.sqrt(2)),
                          ])
def test_l2(x: Any,
            axis: tuple[int, ...] | int | None,
            result: JaxRealArray,
            ) -> None:
    assert_allclose(normalize('l2', jnp.asarray(x), axis=axis), jnp.asarray(result))


@pytest.mark.parametrize("k", [0.0, 1.0])
def test_divide_where(k: float) -> None:
    s = (5, 3)
    w = jnp.ones(s) * k
    x = jnp.arange(math.prod(w.shape), dtype='f').reshape(w.shape)
    dummy = jnp.ones_like(w[..., 0])

    def f(w: Array, x: Array, dummy: Array) -> Array:
        total_w = jnp.sum(w, axis=-1)
        return divide_where(dividend=jnp.sum(w * x, axis=-1),
                            divisor=total_w,
                            where=total_w > 0.0,
                            otherwise=dummy)

    y, vjp_f = vjp(partial(f, dummy=dummy), w, x)
    w_bar, x_bar = vjp_f(jnp.ones_like(y))
    assert_allclose(w_bar, k / s[-1] * jnp.tile(jnp.asarray([-1, 0, 1]), (*s[:-1], 1)), atol=1e-3)
    assert_allclose(x_bar, k / s[-1] * jnp.ones(s))



from math import prod

import jax.numpy as jnp
import pytest

from tjax import Projector


@pytest.mark.parametrize("shape", [(3, 5), (5, 3), (7, 2), (9, 4, 1), (3, 11, 0)])
def test_projection(shape: tuple[int, ...]) -> None:
    x = jnp.reshape(jnp.arange(prod(shape), dtype=float), shape)
    projector = Projector(seed=0)
    projected = projector.project(x)
    assert projected.shape[:-1] == shape[:-1]
    assert projected.shape[-1] == min(shape[-1], 2)



from __future__ import annotations

import re
from typing import Any

import pytest

from tjax import assert_tree_allclose, tree_allclose


@pytest.mark.parametrize(("actual", "desired"),
                         [(0.1234567890123, "0\\.123457"),
                          (0.000123456789, "0\\.00012346"),
                          (1.23456789e-8, "1\\.e-08"),
                          (1.23456789e8, "1\\.23457e\\+08")])
def test_testing(capsys: pytest.CaptureFixture[str], actual: float, desired: str) -> None:
    with pytest.raises(AssertionError) as exception:
        assert_tree_allclose(actual, 1e6, rtol=1e-5, atol=1e-8)
    captured = capsys.readouterr()
    assert not captured.out  # Nothing should be printed.
    assert not captured.err
    assertion_string = exception.value.args[0]
    pattern = rf"""
Tree leaves don't match at position 0 with rtol=1e-05 and atol=1e-08\.
Mismatched elements: 1 / 1 \(100%\)
Maximum absolute difference among violations: .+
Maximum relative difference among violations: .+

Actual: {actual}
Desired: 1000000\.0
Test string:
desired = {desired}"""
    assert re.match(pattern, assertion_string)


@pytest.mark.parametrize(("actual", "desired", "result"),
                         [((), (), True),
                          ((1.2, 3.4), (1.2, 3.4), True),
                          ({'a': 1.2}, {'a': 1.2}, True),
                          (1.3, 1.4, False)])
def test_tree_allclose(actual: Any, desired: Any, *, result: bool | None) -> None:
    assert tree_allclose(actual, desired) == result



from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import Any, TypeVar

from .types_ import GraphEdgeKey, GraphNodeKey

T = TypeVar('T', bound="nx.Graph[Any]")


try:
    import networkx as nx
    from flax.nnx.graph import register_graph_node_type
except ImportError:
    msg = "NetworkX or NNX not available"
    def register_graph_as_nnx_node(graph_type: type[Any]) -> None:
        raise RuntimeError(msg)
else:
    from flax.typing import Key

    from .register_jax import flatten_tree

    def register_graph_as_nnx_node(graph_type: type[T]) -> None:  # pyright: ignore
        # flatten: Callable[[Node], tuple[Sequence[tuple[Key, Leaf]], AuxData]],
        def flatten_graph(graph: T, /
                          ) -> tuple[Sequence[tuple[Key, Any]], None]:
            values, keys = flatten_tree(graph)
            return list(zip(keys, values, strict=True)), None

        # set_key: Callable[[Node, Key, Leaf], None],
        def set_key_graph(graph: T, key: Key, value: Any, /) -> None:
            if not isinstance(value, dict):
                raise TypeError
            d: dict[str, Any]
            match key:
                case GraphNodeKey(name):
                    if not graph.has_node(name):
                        graph.add_node(name, **value)
                        return
                    d = graph.nodes[name]
                case GraphEdgeKey(source, target):
                    if not graph.has_edge(source, target):
                        graph.add_edge(source, target, **value)
                        return
                    d = graph.edges[source, target]
                case _:
                    raise TypeError
            d.clear()
            d.update(value)

        # pop_key: Callable[[Node, Key], Leaf],
        def pop_key_graph(graph: T, key: Key, /) -> Any:
            match key:
                case GraphNodeKey(name):
                    retval = graph.nodes[name]
                    graph.remove_node(name)
                    return retval
                case GraphEdgeKey(source, target):
                    retval = graph.edges[source, target]
                    graph.remove_edge(source, target)
                    return retval
                case _:
                    raise TypeError

        # create_empty: Callable[[AuxData], Node],
        def create_empty_graph(metadata: None, /) -> T:
            del metadata
            return graph_type()

        # clear: Callable[[Node, AuxData], None],
        def clear_graph(graph: T, /) -> None:
            graph.clear()

        # init: tp.Callable[[Node, tp.Iterable[tuple[Key, Leaf]]], None],
        def init_graph(graph: T, it: Iterable[tuple[Key, Any]]) -> None:
            for key, value in it:
                set_key_graph(graph, key, value)

        register_graph_node_type(graph_type,
                                 flatten_graph,
                                 set_key_graph,
                                 pop_key_graph,
                                 create_empty_graph,
                                 clear_graph,
                                 init_graph)



from __future__ import annotations

from collections.abc import Callable
from typing import Any

from jax import tree
from rich.console import Console
from rich.tree import Tree

from .display_generic import display_generic

# Global variables ---------------------------------------------------------------------------------
global_console = Console()


# Functions ----------------------------------------------------------------------------------------
def internal_print_generic(*args: object,
                           raise_on_nan: bool = True,
                           as_leaves: bool = False,
                           console: Console | None = None,
                           ) -> None:
    if console is None:
        console = global_console
    found_nan = False
    root = Tree("", hide_root=True)
    for value in args:
        if as_leaves:
            value = node_to_leaves(value)  # noqa: PLW2901
        _ = root.add(display_generic(value, seen=set()))
        found_nan = found_nan or raise_on_nan and 'nan' in str(root)
    console.print(root)
    if found_nan:
        assert False  # noqa: PT015


def node_to_leaves(tree_: Any, is_leaf: Callable[[Any], bool] | None = None
                   ) -> list[tuple[str, Any]]:
    leaves_and_paths = tree.leaves_with_path(tree_, is_leaf=is_leaf)
    return [("".join(str(x) for x in key_path), leaf)
            for key_path, leaf in leaves_and_paths]



from tjax._src.gradient.aliases import (DPSGD, LARS, LBFGS, SGD, SM3, AdaBelief, AdaDelta,
                                        AdaFactor, AdaGrad, Adam, Adamax, AdamaxW, AdamW, Fromage,
                                        Lamb, Lion, NoisySGD, Novograd, OptimisticGradientDescent,
                                        PolyakSGD, RAdam, RMSProp, Yogi)
from tjax._src.gradient.chain import ChainedGradientState, ChainedGradientTransformation
from tjax._src.gradient.smd import SMDGradient, SMDState
from tjax._src.gradient.transform import (GenericGradientState, GradientState,
                                          GradientTransformation, SecondOrderGradientTransformation,
                                          ThirdOrderGradientTransformation)
from tjax._src.gradient.transforms import (AddDecayedWeights, AddNoise, ApplyEvery, Centralize, Ema,
                                           Scale, ScaleByAdam, ScaleByBelief, ScaleByParamBlockNorm,
                                           ScaleByParamBlockRMS, ScaleByRAdam, ScaleByRms,
                                           ScaleByRss, ScaleBySchedule, ScaleBySM3, ScaleByStddev,
                                           ScaleByTrustRatio, ScaleByYogi, Schedule, Trace)

__all__ = [
    'DPSGD',
    'DPSGD',
    'LARS',
    'LARS',
    'LBFGS',
    'SGD',
    'SGD',
    'SM3',
    'SM3',
    'AdaBelief',
    'AdaDelta',
    'AdaFactor',
    'AdaGrad',
    'Adam',
    'AdamW',
    'Adamax',
    'AdamaxW',
    'AddDecayedWeights',
    'AddNoise',
    'ApplyEvery',
    'Centralize',
    'ChainedGradientState',
    'ChainedGradientTransformation',
    'Ema',
    'Fromage',
    'GenericGradientState',
    'GradientState',
    'GradientTransformation',
    'Lamb',
    'Lion',
    'NoisySGD',
    'Novograd',
    'OptimisticGradientDescent',
    'PolyakSGD',
    'RAdam',
    'RMSProp',
    'SMDGradient',
    'SMDState',
    'Scale',
    'ScaleByAdam',
    'ScaleByBelief',
    'ScaleByParamBlockNorm',
    'ScaleByParamBlockRMS',
    'ScaleByRAdam',
    'ScaleByRms',
    'ScaleByRss',
    'ScaleBySM3',
    'ScaleBySchedule',
    'ScaleByStddev',
    'ScaleByTrustRatio',
    'ScaleByYogi',
    'Schedule',
    'SecondOrderGradientTransformation',
    'ThirdOrderGradientTransformation',
    'Trace',
    'Yogi',
]



from __future__ import annotations

from collections.abc import Hashable
from dataclasses import dataclass
from typing import Any, TypeAlias, override


@dataclass(frozen=True)
class GraphNodeKey:
    """Struct for use with :func:`jax.tree_util.register_pytree_with_keys`."""
    key: Hashable

    @override
    def __str__(self) -> str:
        return f'.node[{self.key!r}]'

    def __lt__(self, value: GraphNodeKey | GraphEdgeKey, /) -> bool:
        return (not isinstance(value, GraphNodeKey)
                or self.key < value.key)  # type: ignore # pyright: ignore


@dataclass(frozen=True)
class GraphEdgeKey:
    """Struct for use with :func:`jax.tree_util.register_pytree_with_keys`."""
    source: Hashable
    target: Hashable

    @override
    def __str__(self) -> str:
        return f'.edge[{self.source!r}, {self.target!r}]'

    def __lt__(self, value: GraphNodeKey | GraphEdgeKey, /) -> bool:
        return isinstance(value, GraphEdgeKey) and (
                (self.source, self.target) < (value.source, value.target))


@dataclass(frozen=True)
class UndirectedGraphEdgeKey(GraphEdgeKey):
    @override
    def __hash__(self) -> int:
        return hash(self.source) ^ hash(self.target)

    @override
    def __lt__(self, value: GraphNodeKey | GraphEdgeKey, /) -> bool:
        return isinstance(value, GraphEdgeKey) and (
                sorted((self.source, self.target))  # type: ignore # pyright: ignore
                < sorted((value.source, value.target)))  # type: ignore # pyright: ignore


GraphAuxData: TypeAlias = list[GraphNodeKey | GraphEdgeKey]  # The auxilliary data for Jax.
GraphData: TypeAlias = dict[str, Any]  # What networkx stores in the graph.



from io import StringIO
from textwrap import dedent
from typing import Any

import jax.numpy as jnp
import jax.random as jr
import numpy as np
from jax import Array, enable_custom_prng, jit, tree, vmap
from pytest import CaptureFixture
from rich.console import Console

from tjax import KeyArray, RealArray, print_generic
from tjax.dataclasses import dataclass, field


def verify(actual: str, desired: str) -> None:
    actual = "\n".join(x.rstrip() for x in actual.strip().split('\n'))
    assert actual == dedent(desired).strip()


def test_numpy_display(capsys: CaptureFixture[str],
                       console: Console) -> None:
    print_generic(np.reshape(np.arange(6.0), (3, 2)), console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           NumPy Array (3, 2) float64
           └──  0.0000 │ 1.0000
                2.0000 │ 3.0000
                4.0000 │ 5.0000
           """)


def test_numpy_display_big(capsys: CaptureFixture[str],
                           console: Console) -> None:
    print_generic(np.reshape(np.arange(30.0), (15, 2)), console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           NumPy Array (15, 2) float64
           ├── mean=14.5
           └── deviation=8.65544144839919
           """)


def test_jax_numpy_display(capsys: CaptureFixture[str],
                           console: Console) -> None:
    print_generic(jnp.ones(3), console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           Jax Array (3,) float64
           └──  1.0000 │ 1.0000 │ 1.0000
           """)


def test_jit_display(capsys: CaptureFixture[str],
                     console: Console) -> None:
    @jit
    def f(x: RealArray) -> RealArray:
        print_generic(x, console=console)
        return x
    f(jnp.asarray(1.0))
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           Jax Array () float64
           └── 1.0000
           """)


def test_vmap_display(capsys: CaptureFixture[str],
                      console: Console) -> None:
    """Like test_batch_display, but uses print_generic to get the array."""
    @jit
    def f(x: RealArray) -> RealArray:
        print_generic(x, console=console)
        return x
    vmap(vmap(f, in_axes=2), in_axes=1)(jnp.ones((3, 4, 5, 6)))
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    s = dedent("""
               Jax Array (3, 5) float64
               └──  1.0000 │ 1.0000 │ 1.0000 │ 1.0000 │ 1.0000
                    1.0000 │ 1.0000 │ 1.0000 │ 1.0000 │ 1.0000
                    1.0000 │ 1.0000 │ 1.0000 │ 1.0000 │ 1.0000""")
    verify(captured, s * 24)


def test_immediate_key_display(capsys: CaptureFixture[str],
                               console: Console) -> None:
    with enable_custom_prng():
        k = jr.key(123)

        @jit
        def f(x: KeyArray) -> None:
            print_generic(x, immediate=True, console=console)

        f(k)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           JaxKey[dataclass]
           └── key_data=Jax Array (2,) uint32
           """)


def test_key_display(capsys: CaptureFixture[str],
                    console: Console) -> None:
    with enable_custom_prng():
        k = jr.key(123)

        @jit
        def f(x: KeyArray) -> None:
            print_generic(x, console=console)

        f(k)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           JaxKey[dataclass]
           └── key_data=Jax Array (2,) uint32
               └──  0 │ 123
           """)


def test_dict_display(capsys: CaptureFixture[str],
                      console: Console) -> None:
    print_generic({'cat': 5,
                   'mouse': {'dog': 3,
                             'sheep': 4}},
                  console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           dict
           ├── cat=5
           └── mouse=dict
               ├── dog=3
               └── sheep=4
           """)


def test_dataclass(capsys: CaptureFixture[str],
                   console: Console) -> None:
    @dataclass
    class C:
        x: int
        y: int

    @dataclass
    class D:
        c: C
        d: C

    print_generic(D(C(1, 2), C(3, 4)),
                  console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           D[dataclass]
           ├── c=C[dataclass]
           │   ├── x=1
           │   └── y=2
           └── d=C[dataclass]
               ├── x=3
               └── y=4
           """)


def test_dataclass_as_leaves(capsys: CaptureFixture[str],
                             console: Console) -> None:
    @dataclass
    class C:
        x: int
        y: int

    @dataclass
    class D:
        c: C
        d: C

    print_generic(D(C(1, 2), C(3, 4)),
                  console=console,
                  as_leaves=True)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           list
           ├── tuple
           │   ├── ".c.x"
           │   └── 1
           ├── tuple
           │   ├── ".c.y"
           │   └── 2
           ├── tuple
           │   ├── ".d.x"
           │   └── 3
           └── tuple
               ├── ".d.y"
               └── 4
           """)


def test_pytreedef(capsys: CaptureFixture[str],
                   console: Console) -> None:
    @dataclass
    class C:
        x: int
        y: int

    _, tree_def = tree.flatten(C(1, 2))

    print_generic(tree_def, immediate=True, console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           f"""
           PyTreeDef
           └── hash={hash(tree_def)}
           """)


def test_seen_array(capsys: CaptureFixture[str],
                    console: Console) -> None:
    @dataclass
    class C:
        x: Array
        y: Array

    z = jnp.zeros(2)
    tree_def = C(z, z)

    print_generic(tree_def, immediate=True, console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           C[dataclass]
           ├── x=Jax Array (2,) float64
           │   └──  0.0000 │ 0.0000
           └── y=Jax Array (2,) float64
               └──  0.0000 │ 0.0000
           """)


if __name__ == "__main__":
    @dataclass
    class Triplet:
        x: Any
        y: Any
        z: Any = field(static=True)

    a = Triplet(np.reshape(np.arange(6.0), (3, 2)),
                np.reshape(np.arange(30.0), (15, 2)),
                Triplet)

    @jit
    def g(x: RealArray) -> None:
        print_generic(Triplet({'abc': Triplet(a,
                                              x,
                                              2)},
                              a,
                              'blah'))
        print_generic(Triplet({'abc': Triplet(a,
                                                     x,
                                                     2)},
                                     a,
                                     'blah'))
    vmap(vmap(g, in_axes=2), in_axes=1)(jnp.ones((3, 4, 5, 6)))
